/*
  Copyright 2003 Laurent Wacrenier

  This file is part of clamd-stream-client.

  clamd-stream-client is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  clamd-stream-client is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with PLL popd; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

static char const rcsid[] =
"$Id: clamd-stream-client.c,v 1.2 2006/03/16 19:04:54 lwa Exp $";

static char const revision[] = "$Revision: 1.2 $";


#include <sys/types.h>
#include <sys/socket.h>

#include <arpa/inet.h>
#include <netinet/in.h>

#include <netdb.h>
#include <sysexits.h>
#include <unistd.h>
#include <fcntl.h>

#include <errno.h>
#include <stdlib.h>
#include <strings.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>

#ifndef DEFAULT_SERVER 
#define DEFAULT_SERVER "127.0.0.1"
#endif

#ifndef DEFAULT_PORT
#define DEFAULT_PORT   "3310"
#endif

int failsafe = 0;

void sigalarm(int sig) {
  fprintf(stderr, "Timeout\n");
  exit(failsafe ? 0 : EX_TEMPFAIL);
}

void usage(char *me) {
  fprintf(stderr, 
	  "usage: %s [-d host] [-p port] [-q[q]] [-f] [-t timeout]\n", 
	  me);
  fprintf(stderr, 
	  "\n"
	  "Send a file in stdin to a clamd server, output the virus name if\n"
	  "found then returns a sendmail exit code (see sysexits(3))\n"
	  "\n"
	  "   -d host  name of Clamd server host (default "DEFAULT_SERVER")\n"
	  "   -p port  port of Clamd server (default "DEFAULT_PORT")\n"
	  "   -q       don't display output\n"
	  "   -qq      don't display output or errors\n"
	  "   -f       return non zero if and only if a virus is found\n"
	  "   -t sec   limit of time this program runs\n"
	  );
  fprintf(stderr, "\n%s\n", revision);
  fprintf(stderr, "Author: Laurent Wacrenier <lwa@teaser.fr>\n");
  /* add your name here if you patch */
}

int check(char *server, char *port) {
  int s;
  struct sockaddr_in sa_in;
  char buf[1024];
  int buf_len= 0;
  int ss;
  char *buf_c;
  char *port_s;
  struct addrinfo hints, *res;
  int err;


  memset(&hints, 0, sizeof(hints));
  hints.ai_family = PF_INET;
  hints.ai_socktype = SOCK_STREAM;

  if ((err = getaddrinfo(server, port, &hints, &res)) != 0) {
    fprintf(stderr, "getaddrinfo(%s:%s): %s\n",
	    server, port, gai_strerror(err));
    return EX_TEMPFAIL;
  }

  while(res) {

    s = socket(PF_INET, SOCK_STREAM, 0);
    if (s == -1) {
      fprintf(stderr, "socket(): %s\n", strerror(errno));
      return EX_OSERR;
    }

    if (connect(s, res->ai_addr, 
		sizeof(struct sockaddr)) == -1) {
      struct sockaddr_in *sa = (struct sockaddr_in *)res->ai_addr;
      fprintf(stderr, "connect(%s:%d): %s\n", 
	      inet_ntoa(sa->sin_addr), (int)ntohs(sa->sin_port),
	      strerror(errno));
      res = res->ai_next;
      close(s);
    } else {
      break;
    }
  }
  if (res == NULL) {
    fprintf(stderr, "unable to connect to %s\n", server);
    return EX_IOERR;
  }

#define COMMAND "STREAM\r\n"
  if (write(s, COMMAND, sizeof(COMMAND)-1) == -1) {
    fprintf(stderr, "write(): %s\n", strerror(errno));
    return EX_IOERR;
  }
  if ((buf_len = read(s, buf, sizeof(buf)-1)) == -1) {
    fprintf(stderr, "read(): %s\n", strerror(errno));
    return EX_IOERR;
  }
  buf[buf_len] = 0;
  if (strncasecmp(buf, "PORT ", sizeof("PORT ") -1) != 0) {
    return EX_PROTOCOL;
  }
  port_s = buf + sizeof("PORT ") -1;
  while(*port_s == ' ') port_s ++;

  memcpy(&sa_in, res->ai_addr, sizeof(sa_in));
  sa_in.sin_port = htons(strtoul(port_s, NULL, 10));
  sa_in.sin_family = AF_INET;


  ss = socket(PF_INET, SOCK_STREAM, 0);
  if (ss == -1) {
    fprintf(stderr, "socket(): %s\n", strerror(errno));
    return EX_OSERR;
  }

  sa_in.sin_port = htons(strtoul(port_s, NULL, 10));
  if (connect(ss, (struct sockaddr *)&sa_in,
              sizeof(struct sockaddr_in)) == -1) {
    fprintf(stderr, "connect2(): %s\n", strerror(errno));
    return EX_OSERR;
  }
  while((buf_len = fread(buf, 1, sizeof(buf), stdin)) > 0) {
    if (write(ss, buf, buf_len) == -1) {
      if (errno == EPIPE)
	break;
      fprintf(stderr, "write2(): %s\n", strerror(errno));
      return EX_IOERR;
    }
  }
  if (buf_len == 0 && ferror(stdin)) {
    fprintf(stderr, "read2(): %s\n", strerror(errno));
    return EX_IOERR;
  }
  close(ss);

  if ((buf_len = read(s, buf, sizeof(buf)-1)) == -1) { 
    fprintf(stderr, "read3(): %s\n", strerror(errno));
    return EX_IOERR;
  }
  buf[buf_len] = 0;

  /*  fwrite(buf, 1, buf_len, stderr); */
  buf_c = buf + buf_len;
  while(*buf_c=='\r' || *buf_c == '\r' || *buf_c == ' ')
    *buf_c --;
  if (buf_c - buf >= sizeof(" FOUND") && 
      strncasecmp(buf_c - sizeof(" FOUND"), " FOUND", 
		  sizeof(" FOUND")-1) == 0) {
    char *buf_s = buf;
    buf_c -= sizeof(" FOUND");
    if (strncasecmp(buf_s, "stream:", sizeof("stream:")-1) == 0) {
      buf_s += sizeof("stream:")-1;
      while(*buf_s == ' ') 
	buf_s ++;
      fprintf(stdout, "%.*s\n", (int)(buf_c - buf_s), buf_s);
    }
    return EX_DATAERR;
  }

  return 0;
}

int main(int argc, char **argv) {
  int ch;
  char *me = argv[0];
  char *server = DEFAULT_SERVER;
  char *port = DEFAULT_PORT;
  int ret;
  int quiet = 0;
  int timeout = 0;

  while ((ch = getopt(argc, argv, "qfd:p:t:h")) != -1)
    switch (ch) {
    case 'q':
      quiet++;
      break;
    case 'f':
      failsafe = 1;
      break;
    case 'd':
      server = optarg;
      break;
    case 'p':
      port = optarg;
      break;
    case 't':
      timeout = strtoul(optarg, NULL, 10);
      break;
    case 'h':
    case '?':
    default:
      usage(me);
      return EX_USAGE;
    }
  argc -= optind;
  argv += optind;
  if (quiet > 0) {
    int fd;
    fd = open("/dev/null", O_WRONLY);
    if (fd > 0 && fd != 2) {
      dup2(fd, 2);
      close(fd);
    }
    if (quiet > 1) {
      dup2(2, 1);
    }
  }

  if (timeout) signal(SIGALRM,sigalarm);

  if (timeout) alarm(timeout);
  ret = check(server, port);
  if (timeout) alarm(0);

  if (failsafe && ret != EX_DATAERR)
    return 0;
  return ret;
}
